#ifndef FORMULA_H
#define FORMULA_H

#include <string>
#include <sstream>
#include <set>
#include <iostream>
#include <typeinfo>
#include <algorithm>

#include "tableau.h"

//==============================================================================
class Formula {
public:
	friend class Binary;
	friend class Conjunction;
	friend class Disjunction;
	friend class Next;
	friend class Until;
	friend class Release;
	
	friend std::ostream & operator<<(std::ostream & o, const Formula & f);

	inline std::string getName(void) const;

	inline FormulaSet elementarySubformulas(void) const;

	inline virtual bool equals(const Formula * f) const;
	bool isInState(const State * s) const;

protected:
	virtual void printTo(std::ostream & o) const = 0;

public:
	virtual const Formula * clone(void) const = 0;
	virtual StateSet sat(const Tableau & t) const = 0;

protected:
	virtual FormulaSet nonUniqueElementarySubformulas(void) const = 0;
};
//==============================================================================
inline std::string Formula::getName(void) const {
	std::stringstream s;
	this->printTo(s);
	return s.str();
}
//==============================================================================
inline bool Formula::equals(const Formula * f) const { 
	return typeid(*this)==typeid(*f);
}
//==============================================================================
inline FormulaSet Formula::elementarySubformulas(void) const {
	FormulaSet result(nonUniqueElementarySubformulas());
		
	for(FormulaSet::const_iterator i=result.begin(); 
		i!=result.end(); ++i) {
		FormulaSet::iterator j=i;
		++j;
			
		while(j!=result.end()) {
			if((*i)->equals(*j)) {
				const Formula * key = *j;
				++j;
				delete key;
				result.erase(key);
			} else {
				++j;
			}
		}	
	}
	
	return result;
}
//==============================================================================
inline bool Formula::isInState(const State * s) const {
	const FormulaSet & fs(*s);
	
	for(FormulaSet::const_iterator i=fs.begin(); i!=fs.end(); ++i) {
		if(this->equals(*i))
			return true;		
	}
	return false;
}
//=============================================================================	
class False: public Formula {
public:
	virtual void printTo(std::ostream & o) const;
	virtual const False * clone(void) const;
	virtual StateSet sat(const Tableau & t) const;
	virtual FormulaSet nonUniqueElementarySubformulas(void) const;

	virtual ~False();
};
//=============================================================================	
class True: public Formula {
public:
	virtual void printTo(std::ostream & o) const;
	virtual const True * clone(void) const;
	virtual StateSet sat(const Tableau & t) const;
	virtual FormulaSet nonUniqueElementarySubformulas(void) const;

	virtual ~True();
};
//=============================================================================	
class AtomicFormula: public Formula {
public:
	virtual bool equals(const Formula * f) const;
	virtual void printTo(std::ostream & o) const;
	virtual const AtomicFormula * clone(void) const;
	virtual StateSet sat(const Tableau & t) const;
	virtual FormulaSet nonUniqueElementarySubformulas(void) const;

private:
	const std::string atom;

public:
	AtomicFormula(std::string a);
	virtual ~AtomicFormula();
	void change(std::string a) const;
};
//=============================================================================	
class Negation: public Formula {
public:
	virtual bool equals(const Formula * f) const;
	virtual void printTo(std::ostream & o) const;
	virtual const Negation * clone(void) const;
	virtual StateSet sat(const Tableau & t) const;
	virtual FormulaSet nonUniqueElementarySubformulas(void) const;

private:
	const AtomicFormula * atom;

public:
	Negation(const AtomicFormula * a);
	virtual ~Negation();
};
//=============================================================================	
class Binary: public Formula {
public:
	virtual FormulaSet nonUniqueElementarySubformulas(void) const;
	virtual bool equals(const Formula * f) const;

	Binary(const Formula * l, const Formula * r);
	virtual ~Binary();
	
	const Formula * left;
	const Formula * right;
};
//=============================================================================	
class Conjunction: public Binary {
public:
	virtual void printTo(std::ostream & o) const;
	virtual const Conjunction * clone(void) const;
	virtual StateSet sat(const Tableau & t) const;
	
	Conjunction(const Formula * l, const Formula * r);
	virtual ~Conjunction();
};
//=============================================================================	
class Disjunction: public Binary {
public:
	virtual void printTo(std::ostream & o) const;
	virtual const Disjunction * clone(void) const;
	virtual StateSet sat(const Tableau & t) const;
	
	Disjunction(const Formula * l, const Formula * r);
	virtual ~Disjunction();
};
//=============================================================================	
class Next: public Formula {
public:
	virtual bool equals(const Formula * ff) const;
	virtual void printTo(std::ostream & o) const;
	virtual const Next * clone(void) const;
	virtual StateSet sat(const Tableau & t) const;
	virtual FormulaSet nonUniqueElementarySubformulas(void) const;

	const Formula * next;

	Next(const Formula * f);
	virtual ~Next();
};
//=============================================================================	
class Until: public Binary {
public:
	virtual void printTo(std::ostream & o) const;
	virtual const Until * clone(void) const;
	virtual StateSet sat(const Tableau & t) const;
	virtual FormulaSet nonUniqueElementarySubformulas(void) const;
	
	Until(const Formula * l, const Formula * r);
	virtual ~Until();
};
//=============================================================================	
class Release: public Binary {
public:
	virtual void printTo(std::ostream & o) const;
	virtual const Release * clone(void) const;
	virtual StateSet sat(const Tableau & t) const;
	virtual FormulaSet nonUniqueElementarySubformulas(void) const;
	
	Release(const Formula * l, const Formula * r);
	virtual ~Release();
};
//==============================================================================
//==============================================================================
inline std::ostream & operator<<(std::ostream & o, const Formula & f) {
	f.printTo(o);
	return o;
}
//==============================================================================
inline void False::printTo(std::ostream & o) const { 
	o << "False"; 
}
//==============================================================================
inline const False * False::clone(void) const { 
	return new False(); 
}
//==============================================================================
inline StateSet False::sat(const Tableau & t) const {
	return StateSet();
}
//==============================================================================
inline FormulaSet False::nonUniqueElementarySubformulas(void) const { 
	return FormulaSet();
}
//==============================================================================
inline False::~False() {}
//==============================================================================
inline void True::printTo(std::ostream & o) const { 
	o << "True"; 
}
//==============================================================================
inline const True * True::clone(void) const { 
	return new True(); 
}
//==============================================================================
inline StateSet True::sat(const Tableau & t) const {
	return t.states;
}
//==============================================================================
inline FormulaSet True::nonUniqueElementarySubformulas(void) const { 
	return FormulaSet();
}
//==============================================================================
inline True::~True() {}
//==============================================================================
inline bool AtomicFormula::equals(const Formula * f) const { 
	return (typeid(*this)==typeid(*f)) &&
		atom==(dynamic_cast<const AtomicFormula *>(f))->atom;
}
//==============================================================================
inline void AtomicFormula::printTo(std::ostream & o) const { 
	o << atom; 
}
//==============================================================================
inline const AtomicFormula * AtomicFormula::clone(void) const {
	return new AtomicFormula(atom);
}
//==============================================================================
inline StateSet AtomicFormula::sat(const Tableau & t) const {
	return t.statesHaving(this);
}
//==============================================================================
inline FormulaSet AtomicFormula::nonUniqueElementarySubformulas(void) const { 
	FormulaSet temp;
	temp.insert(this->clone());
	return temp;
}
//==============================================================================
inline AtomicFormula::AtomicFormula(std::string a) : atom(a) {}
//==============================================================================
inline AtomicFormula::~AtomicFormula() {}
//==============================================================================
inline bool Negation::equals(const Formula * f) const { 
	return (typeid(*this)==typeid(*f)) &&
		(atom->equals((dynamic_cast<const Negation *>(f))->atom));
}
//==============================================================================
inline void Negation::printTo(std::ostream & o) const { 
	o << '~' << *atom;
}
//==============================================================================
inline const Negation * Negation::clone(void) const {
	return new Negation(atom->clone());
}
//==============================================================================
inline StateSet Negation::sat(const Tableau & t) const {
	StateSet A(atom->sat(t));
	
	StateSet result;
	std::insert_iterator<StateSet> ins_it(result, result.begin());
	std::set_difference(
		t.states.begin(), t.states.end(), 
		A.begin(), A.end(), ins_it);	
	
	return result;
}
//==============================================================================
inline FormulaSet Negation::nonUniqueElementarySubformulas(void) 
const { 
	return atom->nonUniqueElementarySubformulas();
}
//==============================================================================
inline Negation::Negation(const AtomicFormula * a) : atom(a) {}
//==============================================================================
inline Negation::~Negation() { delete atom; }
//==============================================================================
inline FormulaSet Binary::nonUniqueElementarySubformulas(void) const { 
	FormulaSet l(left->nonUniqueElementarySubformulas());
	FormulaSet r(right->nonUniqueElementarySubformulas());
	l.insert(r.begin(),r.end());
	return l;
}
//=============================================================================	
inline bool Binary::equals(const Formula * f) const { 
	return (typeid(*this)==typeid(*f)) &&
		left->equals( (dynamic_cast<const Binary *>(f))->left ) &&
		right->equals( (dynamic_cast<const Binary *>(f))->right );
}
//==============================================================================
inline Binary::Binary(const Formula * l, const Formula * r) : left(l), right(r) 
{}
//=============================================================================	
inline Binary::~Binary() { delete left; delete right; }
//=============================================================================	
inline void Conjunction::printTo(std::ostream & o) const { 
	o << "(" << *left << " & " << *right << ")"; 
}
//=============================================================================	
inline const Conjunction * Conjunction::clone(void) const {
	return new Conjunction(left->clone(), right->clone());
}
//=============================================================================	
inline StateSet Conjunction::sat(const Tableau & t) const {
	StateSet result;
	std::insert_iterator<StateSet> ins_it(result, result.begin());
	
	StateSet l(left->sat(t));
	StateSet r(right->sat(t));
	std::set_intersection(l.begin(), l.end(), r.begin(), r.end(), ins_it);
		
	return result;
}
//=============================================================================	
inline Conjunction::Conjunction(const Formula * l, const Formula * r) :
	Binary(l,r) {}
//=============================================================================	
inline Conjunction::~Conjunction() {}
//=============================================================================	
inline void Disjunction::printTo(std::ostream & o) const { 
	o << "(" << *left << " | " << *right << ")"; 
}
//=============================================================================	
inline const Disjunction * Disjunction::clone(void) const {
	return new Disjunction(left->clone(), right->clone());
}
//=============================================================================	
inline StateSet Disjunction::sat(const Tableau & t) const {
	StateSet result;
	std::insert_iterator<StateSet> ins_it(result, result.begin());
	
	StateSet l(left->sat(t));
	StateSet r(right->sat(t));
	std::set_union(l.begin(), l.end(), r.begin(), r.end(), ins_it);
		
	return result;
}
//=============================================================================	
inline Disjunction::Disjunction(const Formula * l, const Formula * r) :
	Binary(l,r) {}
//=============================================================================	
inline Disjunction::~Disjunction() {}
//=============================================================================	
inline bool Next::equals(const Formula * f) const { 
	return (typeid(*this)==typeid(*f)) &&
		next->equals( (dynamic_cast<const Next *>(f))->next ); 
}
//=============================================================================	
inline void Next::printTo(std::ostream & o) const { 
	o << "AX(" << *next << ")"; 
}
//=============================================================================	
inline const Next * Next::clone(void) const {
	return new Next(next->clone());
}
//=============================================================================	
inline StateSet Next::sat(const Tableau & t) const {
	if(!t.isElementaryFormula(this)) {
		std::cout << "\n===Sat called on non-elementary AX formula===\n";
	}
	
	StateSet result;
	
	for(StateSet::const_iterator i=t.states.begin(); i!=t.states.end(); ++i) {
		if(isInState(*i)) 
			result.insert(*i);
	}
	
	return result;
}
//=============================================================================	
inline FormulaSet Next::nonUniqueElementarySubformulas(void) const { 
	FormulaSet result(next->nonUniqueElementarySubformulas());
	result.insert(this->clone());
	return result;
}
//=============================================================================	
inline Next::Next(const Formula * f) : next(f) {}
//=============================================================================	
inline Next::~Next() { delete next; }
//=============================================================================	
inline void Until::printTo(std::ostream & o) const { 
	o << "A[" << *left << " U " << *right << "]";
}
//=============================================================================	
inline const Until * Until::clone(void) const {
	return new Until(left->clone(),right->clone());
}
//=============================================================================	
inline StateSet Until::sat(const Tableau & t) const {
	if(t.isElementaryFormula(this)) {
		StateSet result;
	
		for(StateSet::const_iterator i=t.states.begin(); 
			i!=t.states.end(); ++i) {
			if(this->isInState(*i)) 
				result.insert(*i);
		}
	
		return result;
	} else {
		const Formula * f = new Next(this->clone());
		StateSet A(f->sat(t));
		StateSet L(left->sat(t));

		StateSet B;
		std::insert_iterator<StateSet> ins_it(B, B.begin());
		std::set_intersection(A.begin(), A.end(), L.begin(), L.end(), ins_it);
		
		StateSet R(right->sat(t));
		B.insert(R.begin(), R.end());
		
		const Formula * g = new Next(new False);
		StateSet C(g->sat(t));
		B.insert(C.begin(), C.end());
		
		delete g;
		delete f;
				
		return B;
	}
}
//==============================================================================
inline FormulaSet Until::nonUniqueElementarySubformulas(void) const { 
	FormulaSet result;
	result.insert(new Next(new False()));
	result.insert(new Next(this->clone()));
		
	FormulaSet sub(Binary::nonUniqueElementarySubformulas());
	result.insert(sub.begin(), sub.end());
		
	return result;
}
//=============================================================================	
inline Until::Until(const Formula * l, const Formula * r) : Binary(l,r) {}
//=============================================================================	
inline Until::~Until() {}
//=============================================================================	
inline void Release::printTo(std::ostream & o) const { 
	o << "A[" << *left << " R " << *right << "]";
}
//=============================================================================	
inline const Release * Release::clone(void) const {
	return new Release(left->clone(),right->clone());
}
//=============================================================================	
inline StateSet Release::sat(const Tableau & t) const {
	if(t.isElementaryFormula(this)) {
		StateSet result;
	
		for(StateSet::const_iterator i=t.states.begin(); 
			i!=t.states.end(); ++i) {
			if(this->isInState(*i)) 
				result.insert(*i);
		}
	
		return result;
	} else {
		const Formula * f = new Next(this->clone());
		StateSet A(f->sat(t));
		StateSet L(left->sat(t));
		A.insert(L.begin(), L.end());
		
		StateSet R(right->sat(t));
		StateSet B;
		std::insert_iterator<StateSet> ins_it(B, B.begin());
		std::set_intersection(A.begin(), A.end(), R.begin(), R.end(), ins_it);
		
		const Formula * g = new Next(new False);
		StateSet C(g->sat(t));
		B.insert(C.begin(), C.end());
		
		delete g;
		delete f;
				
		return B;
	}
}
//=============================================================================	
inline FormulaSet Release::nonUniqueElementarySubformulas(void) const { 
	FormulaSet result;
	result.insert(new Next(new False()));
	result.insert(new Next(this->clone()));
	
	FormulaSet sub(Binary::nonUniqueElementarySubformulas());
	result.insert(sub.begin(), sub.end());
	
	return result;
}
//=============================================================================	
inline Release::Release(const Formula * l, const Formula * r) : Binary(l,r) {}
//=============================================================================	
inline Release::~Release() {}
//=============================================================================	
#endif
