#include "model.h"
#include "formula.h"

//=============================================================================	
std::string Model::reformat(const Formula * f) const {
	std::string result;
	std::string s(f->getName());
		
	for(unsigned i = 0; i<s.size(); ++i) {
	switch(s[i]) {
		case ' ':
			result+='_';
			continue;
			
		case '[':
		case ']':
			result+='$';
			continue;

		case '(':
		case ')':
			result+='#';
			continue;		

		case '~':
			result+="_not_";
			continue;		
		
		case '|':
			result+="_or_";
			continue;		

		case '&':
			result+="_and_";
			continue;		
		
		default:
			result+=s[i];
			continue;
	}}
	return result;
}
//==============================================================================
std::ostream & operator<<(std::ostream & o, const FormulaSet & fs) {
	for(FormulaSet::const_iterator ct=fs.begin(); ct!=fs.end(); ++ct) {
		o << *(*ct) << ", ";
	}
	return o;
}
//==============================================================================
std::ostream & operator<<(std::ostream & o, const StateSet & s) {
	for(StateSet::const_iterator ct=s.begin(); ct!=s.end(); ++ct) {
		o << *(*ct) << std::endl;
	}
	return o;
}
//==============================================================================
bool Model::isElementaryFormula(const Formula * f) const {
	return f->isInState(&elementaryFormulas);
}
//==============================================================================
StateSet Model::statesHaving(const Formula * f) const {
	StateSet result;
	
	for(StateSet::const_iterator i=states.begin(); i!=states.end(); ++i) {
		if(f->isInState(*i))
			result.insert(*i);
	}
	
	return result;
}
//==============================================================================
void Model::toDot(std::ostream & o) const {
	o << "digraph tableau {" << std::endl;
	o << "node [fontname=\"Helvetica\"];\n";
//	o << "rankdir=LR;" << std::endl;
//	o << "ratio=;" << std::endl;
//	o << "size = \"1000,1000\";" << std::endl;
//	o << "bgcolor=white;" << std::endl;
//	o << "node [style=filled, fillcolor=green]" << std::endl;

	for(StateSet::const_iterator ci=states.begin(); ci!=states.end(); ++ci) {
		o << "n" << *ci;
		
		o << " [label=\"";
		for(FormulaSet::const_iterator f=(*ci)->begin(); f!=(*ci)->end(); ++f) {
			std::string form = (*f)->getName();
			std::string reformatted = reformat(*f);
			if(reformatted!=form) {
				//o << reformat(*f) << "\\n";
				o << *(*f) << ", " << reformat(*f) << "\\n";
			} else {
				o << *(*f); // << "\\n";
			}			
		}
		o << "\"";		
		
		if(isInSet(*ci, initialStates)) {
			o << ", shape=diamond";
		}
		
//		if(isInSet(*ci, fairStates)) {
//			o << ", fillcolor=red";
//		}
		
		o << "];" << std::endl;

		
// 		TransitionRelation::const_iterator tr=transitionRelation.find(*ci);
// 		for(StateSet::const_iterator successor=((*tr).second)->begin();
// 			successor!=((*tr).second)->end(); ++successor) {
// 				o << "n" << (*tr).first << " -> n" << *successor  << ";";
// 				o << std::endl;
// 			}
// 		
// 		o << std::endl;
		
	}
	
// 	o << std::endl;
// 	
// 	for(TransitionRelation::const_iterator tr=transitionRelation.begin();
// 		tr!=transitionRelation.end(); ++tr) {
// 		
// 		for(StateSet::const_iterator successor=((*tr).second)->begin();
// 			successor!=((*tr).second)->end(); ++successor) {
// 				o << "n" << (*tr).first << " -> n" << *successor  << ";";
// 				o << std::endl;
// 			}
// 		
// 		o << std::endl;
// 	}
	
	for(TransitionRelation::const_iterator tr=transitionRelation.begin();
		tr!=transitionRelation.end(); ++tr) {
		
		for(StateSet::const_iterator successor=states.begin();
			successor!=states.end(); ++successor) {
				if(((*tr).second)->find(*successor)==((*tr).second)->end()) {
					o << "n" << (*tr).first << " -> n" << *successor  << ";";
					o << std::endl;
				}
			}
		
		o << std::endl;
	}

	o << "}" << std::endl;
}
//==============================================================================
inline std::string Model::stateFormula(const State * s, bool next) const {
	std::string result;
	FormulaSet::const_iterator begin = elementaryFormulas.begin();
	FormulaSet::const_iterator end = elementaryFormulas.end();

	for(FormulaSet::const_iterator i=begin; i!=end; ++i) {
	
		if(i!=begin && i!=end) {
			result += "&";
		}

		if(!isInSet(*i, *s)) {
			result += "!";
		}
		
		if(next) {
			result += "next(";
		}
		result += reformat(*i);
		if(next) {
			result += ")";
		}
	}
	return result;
}
//==============================================================================
void Model::toSMV(std::ostream & o, std::string spec) const {
	o << "MODULE main\n";

	o << "VAR\n";
	for(FormulaSet::const_iterator i=elementaryFormulas.begin();
		i!=elementaryFormulas.end(); ++i) {
		
		o << '\t' << reformat(*i) << " : boolean;\n";
	}
	o << '\n';
	
	o << "TRANS\n";
	for(StateSet::const_iterator i=states.begin(); i!=states.end(); ++i) {
	for(StateSet::const_iterator j=states.begin(); j!=states.end(); ++j) {
		if(isInSet(*j, *((transitionRelation.find(*i))->second))){
			o << '\t' << stateFormula(*i, false) << " & ";
			o << stateFormula(*j, true) << "|\n";
		}
	}}
	o << "\t0 \n";


/*	StateSet reachable(computeReachable());

	for(StateSet::const_iterator i=states.begin(); i!=states.end(); ++i) {
		if(!isInSet(*i, reachable)) {
			continue;
		}
		o << "TRANS\t";
		o << "!(" << stateFormula(*i, false) << "&(";
	
		for(StateSet::const_iterator j=states.begin(); j!=states.end(); ++j) {
			if(!isInSet(*j, *((transitionRelation.find(*i))->second))){
				o << "(" << stateFormula(*j, true) << ")|";
			}
		}
		o << "0))\n";
	
	}
	o << "\n";
*/
	o << "INIT\n";
	for(StateSet::const_iterator j=states.begin(); j!=states.end(); ++j) {
		if(isInSet(*j, initialStates)) {
			o << '\t' << stateFormula(*j, false) << "|\n";
		}
	}
	o << "\t0 \n";
	
	for(FairnessConstraints::const_iterator i=fairnessConstraints.begin();
		i!=fairnessConstraints.end(); ++i) {
		o << "FAIRNESS\n";
		
		for(StateSet::const_iterator j=states.begin(); j!=states.end(); ++j) {
			if(isInSet(*j, *(*i))) {
				o << '\t' << stateFormula(*j, false) << "|\n";
			}
		}
		o << "\t0 \n";
	}
	
	o << "SPEC " << spec << '\n';
}
//==============================================================================
void Model::removeTransitionsTo(const State *s) {
	for(StateSet::const_iterator i=states.begin(); i!=states.end(); ++i) {
		if(isInSet(s, *(transitionRelation[*i]))) {
			transitionRelation[*i]->erase(s);
		}	
	}
}
//==============================================================================
void Model::removeState(const State * s) {
	if(!isInSet(s, states)) {
		return;
	}
	
	removeTransitionsTo(s);
		
	if(isInSet(s, initialStates)) {
		initialStates.erase(s);
	}

	for(std::set<StateSet *>::iterator f=fairnessConstraints.begin();
		f!=fairnessConstraints.end();
		++f) {
		if(isInSet(s, *(*f))) {
			(*f)->erase(s);
		}
	}

	delete transitionRelation[s];
	transitionRelation.erase(s);
		
	states.erase(s);
	delete s;
}
//==============================================================================
void Model::removeNonTotalStates(void) {
	do {
		StateSet::const_iterator i=states.begin();
		while(i!=states.end() && !(transitionRelation[*i]->empty())) { ++i; }
		
		if(i==states.end()) { break; }

		removeState(*i);		
	} while(true);
}
//==============================================================================
Model::~Model() {
	for(std::set<StateSet *>::const_iterator 
		i=fairnessConstraints.begin();
		i!=fairnessConstraints.end();
		++i) {
		delete (*i);
	}

	for(TransitionRelation::iterator i=transitionRelation.begin();
		i!=transitionRelation.end(); ++i) {
		delete (*i).second;
	}

	for(StateSet::iterator i=states.begin(); i!=states.end(); ++i) {
		delete *i;
	}

	for(FormulaSet::iterator i=elementaryFormulas.begin(); 
		i!=elementaryFormulas.end(); ++i) {
		delete *i;
	}
}
//==============================================================================
StateSet Model::computeReachable(void) const {
	StateSet r = initialStates;
	bool more = true;
	
	do {
		StateSet temp;
		for(StateSet::const_iterator i=r.begin(); i!=r.end(); ++i) {
			temp.insert(
				(transitionRelation.find(*i)->second)->begin(),
				(transitionRelation.find(*i)->second)->end());
//				transitionRelation[*i]->begin(),
//				transitionRelation[*i]->end());
		}
		unsigned before = r.size();
		r.insert(temp.begin(), temp.end());
		unsigned after = r.size();
		more = (after-before)>0;
	} while(more);
	
	return r;
}
//==============================================================================
void Model::removeUnreachable(void) {
	StateSet r = computeReachable();
	StateSet s = states;
		
	for(StateSet::const_iterator i=s.begin(); i!=s.end(); ++i) {
		if(!isInSet(*i, r)) {
			removeState(*i);
		}
	}

}
//==============================================================================
bool Model::increase(std::vector<unsigned> & bitVector) {
	unsigned i = 0;
	for(; i<bitVector.size(); ++i) {
		if(bitVector[i]==0) {
			bitVector[i]++;
			break;
		} else {
			bitVector[i]=0;
		}
	}
	return(i==bitVector.size());
}
//==============================================================================
Model * Model::composeWith(const Model & m) const {
	Model * pr = new Model;
/*	Model & product = *pr;
	
	FormulaSet this_formulas;
	FormulaSet m_formulas;

	// clone formulas	
	for(FormulaSet::const_iterator f=elementaryFormulas.begin(); 
		f!=elementaryFormulas.end(); ++f) {
		
		this_formulas.insert((*f)->clone());
	}
	
	for(FormulaSet::const_iterator f=m.elementaryFormulas.begin(); 
		f!=m.elementaryFormulas.end(); ++f) {
		
		bool found = false;
		
		for(FormulaSet::const_iterator g=elementaryFormulas.begin(); 
			(g!=elementaryFormulas.end()) && !found; ++g) {
		
			found = ((*f)->getName()==(*g)->getName());
		}
		
		if(!found) {
			m_formulas.insert((*f)->clone());
		}
	}
	
	product.insert(this_formulas.begin(), this_formulas.end());
	product.insert(m_formulas.begin(), m_formulas.end());
	
	
	// states -> the nasty way
	std::vector<unsigned> bitVector(product.elementaryFormulas.size(), 0);
	
	do {
		State * s = new State();
		
		std::vector<unsigned>::const_iterator i =  bitVector.begin();
		FormulaSet::const_iterator j = product.elementaryFormulas.begin();
		
		for(; i!=bitVector.end(); ++i, ++j) {
			if(*i>0) {
				s->insert(*j);
			}
		}

		product.states.insert(s);
		
	} while(!increase(bitVector));
	
	// initial states
	for(StateSet::const_iterator i=product.states.begin(); 
		i!=product.states.end(); ++i) {
		
		
	}
*/
	return pr;
}
//==============================================================================
