#include "tableau.h"
#include "formula.h"

//==============================================================================
Tableau::Tableau(const Formula *f)
{
	buildElementaryFormulas(f);
	buildStates();
	initialStates = f->sat(*this);
	buildFairness();
	constructTransitions();
//	removeUnreachable();
//	removeNonTotalStates();
}
//==============================================================================
void Tableau::buildElementaryFormulas(const Formula *f) {
	elementaryFormulas = f->elementarySubformulas();
	
	const Formula * at = new AtomicFormula("aa");
	unsigned j = 0;
	
	for(FormulaSet::const_iterator i=elementaryFormulas.begin();
		i!=elementaryFormulas.end(); ++i) {
		
		if(typeid(*(*i))==typeid(*at)) {
			reformatter[*i] = (*i)->getName();
		} else {
			std::stringstream s;
			s << 'p' << j;
			reformatter[*i] = s.str();
			++j;
		}
	
	}
	
	delete at;
}
//==============================================================================
void Tableau::buildStates(void) {
	std::vector<unsigned> bitVector(elementaryFormulas.size(), 0);
	
	do {
		State * s = new State();
		
		std::vector<unsigned>::const_iterator i =  bitVector.begin();
		FormulaSet::const_iterator j = elementaryFormulas.begin();
		
		for(; i!=bitVector.end(); ++i, ++j) {
			if(*i>0) {
				s->insert(*j);
			}
		}

		states.insert(s);
		
	} while(!increase(bitVector));
}
//==============================================================================
void Tableau::constructTransitions(void) {
	// first find all elementary formulas of the for AX g
	FormulaSet nextFormulas;
	const Formula * someNextFormula = new Next(new False);
	
	for(FormulaSet::const_iterator f=elementaryFormulas.begin(); 
		f!=elementaryFormulas.end(); ++f) {
		if(typeid(*someNextFormula)==typeid(*(*f)))
			nextFormulas.insert(*f);
	}
	
	delete someNextFormula;
	
	// now calculate for each el.form. AX g 
	// sat(AX g) and sat(g)
	
	typedef std::pair<const StateSet *, const StateSet *> PairOfSets;
	typedef std::set<const PairOfSets *> SatSets;
	SatSets satSets;
	
	for(FormulaSet::const_iterator f=nextFormulas.begin(); 
		f!=nextFormulas.end(); ++f) {
		
		const Formula * g = (dynamic_cast<const Next *>(*f))->next;

		const StateSet * F = new StateSet((*f)->sat(*this));
		const StateSet * G = new StateSet(g->sat(*this));
		const PairOfSets * p = new PairOfSets(F, G);
		satSets.insert(p);
	}
	
	
	// ctr tr-rel
	for(StateSet::const_iterator s1=states.begin(); s1!=states.end(); ++s1) {
		StateSet * s1_successors = new StateSet;
	
		for(StateSet::const_iterator s2=states.begin(); 
			s2!=states.end(); ++s2) {

			bool all = true;
		
			for(SatSets::const_iterator f=satSets.begin(); f!=satSets.end();
				++f) {
		
				if(!isInSet(*s1, *((*f)->first))) {
					continue;
				}
		
				if(!isInSet(*s2, *((*f)->second))) {
					all = false;
					break;
				}
			}
			
			if(all) {
				s1_successors->insert(*s2);
			}
		}
		
		transitionRelation[*s1] = s1_successors;
	}
	
	// clean up
	for(SatSets::iterator f=satSets.begin(); f!=satSets.end(); ++f) {
		delete (*f)->first;
		delete (*f)->second;
		delete (*f);
	}
}
//==============================================================================
void Tableau::buildFairness(void) {
	const Next * form = new Next(new Until (new False, new False));
	FormulaSet fs;
	
	for(FormulaSet::const_iterator el=elementaryFormulas.begin();
		el!=elementaryFormulas.end(); ++el) {
		
		if( typeid(*form)==typeid(*(*el)) &&
			typeid(*(form->next))==
				typeid(*((dynamic_cast<const Next *>(*el))->next))) {
		
			fs.insert(*el);
		}
	}
	
	for(FormulaSet::const_iterator el2=fs.begin(); el2!=fs.end(); ++el2) {
		// temp1 = sat(AX A[left U right])
		StateSet temp1((*el2)->sat(*this));
		StateSet temp2;
		std::insert_iterator<StateSet> ins_it(temp2, temp2.begin());
		
		// temp2 = States - sat(AX A[left U right])
		std::set_difference(
			states.begin(), states.end(), temp1.begin(), temp1.end(), ins_it);
		
		// temp3 = sat(right)
		StateSet * temp3 = new StateSet(
			((dynamic_cast<const Until *>(
				((dynamic_cast<const Next *>(*el2))->next)
			))->right)->sat(*this)
		);
		
		// temp3 = (States - sat(AX A[left U right])) union sat(right)
		temp3->insert(temp2.begin(), temp2.end());
		
		fairnessConstraints.insert(temp3);
	}

	delete form;
}
//==============================================================================
std::string Tableau::reformat(const Formula * f) const {
	return (reformatter.find(f))->second;
}
//==============================================================================

